/* CSA.cpp
   Copyright 2003, V. Makinen, all rights reserved.

   This file contains an implementation of the Compact Suffix Array
   index structure [1,2,3].

   This software may be used freely for any non-commercial purpose. However, when distributed,
   the original source must be clearly stated, and, when the source code is
   distributed, the copyright notice must be retained and any alterations in
   the code must be clearly marked. No warranty is given regarding the quality
   of this software.

   [1] Veli Mkinen. Compact Suffix Array. In Proc. 11th Annual Symposium on
       Combinatorial Pattern Matching (CPM 2000), Springer-Verlag LNCS 1848,
       pp. 305-319, Montral, June 2000.
   [2] Veli Mkinen. Trade off Between Compression and Search Times in Compact
       Suffix Array. In Proc. 3rd Workshop on Algorithm Engineering and
       Experimentation (ALENEX 2001), Springer-Verlag LNCS 2153, pp. 189-201,
       Washington D.C., January 2001.
   [3] Veli Mkinen. Compact Suffix Array - A Space-Efficient Full-Text Index.
       Fundamenta Infomaticae, Special Issue - Computing Patterns in Strings, 
       56(1-2):191-210,2003.
*/

//---------------------------------------------------------------------------

#include "TSA.h"
#include "CSA.h"

#define maxx(A,B) ((A) > (B) ? (A) : (B))
#define minn(A,B) ((A) < (B) ? (A) : (B))
//---------------------------------------------------------------------------

typedef struct {
   unsigned signbit : 1;
   unsigned LCP : 31;
} lcpmask;

typedef struct {
   unsigned signbit : 1;
   unsigned pos : 31;
} posmask;

typedef struct {
   unsigned jump : 16;
   unsigned depth : 16;
} jumpmask;

typedef union {
   lcpmask as_lcp;
   jumpmask as_jump;
} lcpjumpmask;

typedef struct SAElement {
    posmask a;
} TSAElement;

typedef struct LCPElement {
  lcpjumpmask a;
} TLCPElement;

typedef struct {
   unsigned int lLCP;
   unsigned int rLCP;
} TLCPPair;

unsigned n,m;
unsigned w=32;
bool print=true;
int report=1;
long tuc, maxtucindex=0, maxtuc=0, maxtucsize=0;
unsigned maxdepth, maxLCP;
double sumdepth;
unsigned *depthcount = (unsigned *)malloc(10000*sizeof(unsigned));
unsigned operations;
unsigned C=15;
unsigned bitsn=0;
unsigned savemem=0;
unsigned SuffixArrayCompactSize = 0;
char* text;
char* pattern;
char* filename;
TSA *SA,*CSA;
BitSequence * Lmap;		// BitSequence RRR that delimit where start each position of CSA in SA

char *SubString(char * A,unsigned start, unsigned count)
{
   char * Result = (char*)malloc((count+1)*sizeof(char));
   unsigned i;
   i=start;
   unsigned j=i;
   while (i < start+count && A[i] != '\0') {
      if (A[i]!= '\n') {
         Result[j-start] = A[i];
         j++;
      }
      i++;
   }
   Result[i-start] = '\0';
   return Result;
}

int ComparePos(unsigned pos, char *pattern)
{
   unsigned j=0;
   while (pos+j < n && j < m && text[pos+j]==pattern[j])
      j++;
   if (j >= m)
      return 0;
   else if (pos+j >= n)
      return +1;
   else if (text[pos+j]>pattern[j])
      return -1;
   else
      return +1;

}

int LCP(char *a, int apos, char *b, int bpos)
{
   int i=0;
   while (a[apos+i]!='\0' && b[bpos+i] != '\0' && a[apos+i]==b[bpos+i])
       i++;
   return i;
}


inline int CompareSuffixes(char *S, unsigned x, unsigned y, unsigned depth)
{
      int inc=0;
      if (x==y) return 0;
      while (x+inc+depth < n && y+inc+depth < n && S[x + inc+depth] == S[y + inc+depth])
         inc++;
      if (x+inc+depth == n)
         return -(inc+1);
      else if (y+inc+depth == n)
         return +(inc+1);
      else
         return S[x+inc+depth] < S[y+inc+depth] ? -(inc+1) : +inc+1;
}

void CalcAdjacentLCPsSA(TSA *SArray)
{
   unsigned i,j;
   unsigned index,x;
   unsigned comparisons=0;
   for (i=0; i<n; i++) {
      assert(SArray->GetPos(i)<n);
      SArray->SetLCP(SArray->GetPos(i),i);
   }
   unsigned lastLCP=0, newLCP=0;
   double sumLCP=0.0, sumLCPv=0.0;
   for (i=0; i<n; i++) {
      index = SArray->GetLCP(i);
      if (index == 0)
         SArray->SetLCP(i,0);
      else {
         if (lastLCP > 0)
            lastLCP--;
         newLCP = 
         abs(CompareSuffixes(text,SArray->GetPos(index),SArray->GetPos(index-1),
                lastLCP));
         SArray->SetLCP(i,lastLCP+newLCP-1);
         comparisons += newLCP;
      }
      lastLCP = SArray->GetLCP(i);
   }
   maxLCP =0;
   for (i=0; i<n; i++) {
      if (SArray->GetLCP(SArray->GetPos(i)) >maxLCP)
         maxLCP = SArray->GetLCP(SArray->GetPos(i));
      sumLCP += SArray->GetLCP(SArray->GetPos(i));
   }
   for (i=0; i<n; i++) {
      sumLCPv += (SArray->GetLCP(SArray->GetPos(i))-sumLCP/n)*(SArray->GetLCP(SArray->GetPos(i))-sumLCP/n);
   }
   unsigned pLCP,cLCP;
   for (i=0; i<n; i++)
      SArray->SetB(i,0);
   for (i=0; i<n; i++) {
      j = SArray->GetPos(i);
      x = i;
      while (SArray->GetB(j)==0) {
         index = j;
         j = SArray->GetPos(index);
         SArray->SetPos(index,x);
         SArray->SetB(index,1);
         x = index;
      }
   }
   for (i=0; i<n; i++)
      SArray->SetB(i,0);
   for (i=0; i<n; i++) {
      j = SArray->GetPos(i);
      x = i;
      pLCP = SArray->GetLCP(x);
      while (SArray->GetB(j)==0) {
         index = j;
         j = SArray->GetPos(index);
         cLCP = SArray->GetLCP(index);
         SArray->SetLCP(index,pLCP);
         pLCP = cLCP;
         SArray->SetPos(index,x);
         SArray->SetB(index,1);
         x = index;
      }
   }
   for (i=0; i<n; i++)
      SArray->SetB(i,0);
   if (print) {
      printf("comparisons=%d <= 2*n=%d\n",comparisons,2*n);
      printf("maxLCP=%d\n",maxLCP);
      printf("avgLCP=%f\n",sumLCP/n);
      printf("sqrt(variance)=%f\n",sqrt(sumLCPv/n));
      fflush(stdout);
   }
}


TLCPPair CalcLCPsFromAdjacentLCPs(TSA *CArray,unsigned l, unsigned r)
{
   unsigned m = (l+r) / 2;
   unsigned i;
   TLCPPair lp,rp,tp;
   tp.lLCP = 0;
   tp.rLCP = 0;
   if (CArray->GetB(m))
      m--;
   assert(!CArray->GetB(m));
   if (m==l+1)
      tp.lLCP = CArray->GetLCP(m);
   else if (CArray->GetB((l+m)/2) && (l+m)/2 <= l+1) {
      tp.lLCP = CArray->GetLCP(m);
      for (i=l+1;i<m;i++)
         if (!CArray->GetB(i) && CArray->GetLCP(i) < tp.lLCP)
            tp.lLCP = CArray->GetLCP(i);
   }
   else {
      lp = CalcLCPsFromAdjacentLCPs(CArray,l,m);
      tp.lLCP = minn(lp.lLCP,lp.rLCP);
   }
   if (m==r-1)
      tp.rLCP = CArray->GetLCP(r);
   else if (CArray->GetB((m+r)/2) && (m+r)/2 <= m+1) {
      tp.rLCP = CArray->GetLCP(r);
      for (i=m+1;i<r;i++)
         if (!CArray->GetB(i) && CArray->GetLCP(i) < tp.rLCP)
            tp.rLCP = CArray->GetLCP(i);
   }
   else {
      rp = CalcLCPsFromAdjacentLCPs(CArray,m,r);
      tp.rLCP = minn(rp.lLCP,rp.rLCP);
   }
   if (tp.lLCP >= tp.rLCP) {
      CArray->SetLCPl(m);
      CArray->SetLCP(m,tp.lLCP);
   }
   else {
      CArray->SetLCPr(m);
      CArray->SetLCP(m,tp.rLCP);
   }
   if (l==0 && CArray->GetLCPl(0)==0) {
      CArray->SetLCP(0,tp.lLCP);
      CArray->SetLCPr(0);
      CArray->SetLCP(r,tp.rLCP);
      CArray->SetLCPl(r);
   }
   return tp;
}

TLCPPair CalcLCPs(TSA *CArray,unsigned l, unsigned r)
{
   unsigned m = (l+r) / 2;
   unsigned i;
   TLCPPair lp,rp,tp;
   tp.lLCP = 0;
   tp.rLCP = 0;
   if (CArray->GetB(m))
      m--;
   assert(!CArray->GetB(m));
   if (m==l+1 || (CArray->GetB((l+m)/2) && (l+m)/2 <= l+1)) {
      tp.lLCP =
      abs(CompareSuffixes(text,CArray->GetPos(l),CArray->GetPos(m),0))-1;
   }
   else {
      lp = CalcLCPs(CArray,l,m);
      tp.lLCP = minn(lp.lLCP,lp.rLCP);
   }
   if (m==r-1 ||(CArray->GetB((m+r)/2) && (m+r)/2 <= m+1)) {
      tp.rLCP =
      abs(CompareSuffixes(text,CArray->GetPos(m),CArray->GetPos(r),0))-1;
   }
   else {
      rp = CalcLCPs(CArray,m,r);
      tp.rLCP = minn(rp.lLCP,rp.rLCP);
   }
   if (tp.lLCP >= tp.rLCP) {
      CArray->SetLCPl(m);
      CArray->SetLCP(m,minn(tp.lLCP,maxLCP));
   }
   else {
      CArray->SetLCPr(m);
      CArray->SetLCP(m,minn(tp.rLCP,maxLCP));
   }
   if (l==0 && CArray->GetLCPl(0)==0) {
      CArray->SetLCP(0,minn(tp.lLCP,maxLCP));
      CArray->SetLCPr(0);
      CArray->SetLCP(r,minn(tp.rLCP,maxLCP));
      CArray->SetLCPl(r);
   }
   return tp;
}

unsigned SearchFromCSA(char *pattern, TSA *SArray, unsigned length)
{
   unsigned k;
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   unsigned limit;
   if (SArray->GetB(i))
      i--;
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (L<R-1 && i!=L) {
         if (ComparePos(SArray->GetPos(i),pattern) <= 0)
            R = i;
         else
            L = i;
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Lraja=L+1;
         if (L==R-3 && ComparePos(SArray->GetPos(R-1),pattern) > 0)
            Lraja = R;
      }
      else
         Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   if (SArray->GetB(i))
      i--;
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1 && i!=L) {
         if (ComparePos(SArray->GetPos(i),pattern) >= 0)
            L = i;
         else
            R = i;
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Rraja=L+1;
         if (L== R-3 && ComparePos(SArray->GetPos(R-1),pattern) >= 0)
            Rraja = R-1;
      }
      else
         Rraja = L;
   }
   if (Lraja+1 < Rraja && !report) {
      if (print) printf("Pattern (%s) found\n",pattern);
      return 1;
   }   
   else if ((Lraja==Rraja && SArray->GetB(Lraja)) || Lraja < Rraja) {
      TSA *DArray = new TSA();
      DArray->CreatePosArray(SArray->posbits,(Rraja-Lraja+1)*C);
      unsigned linkstart=0;
      unsigned j=0;
      for (i=Lraja; i<=Rraja; i++) {
         if (SArray->GetB(i)) {
            linkstart=0;
            UnCompactLink(SArray,
                       i,
                       0,
                       DArray,
                       j,
                       SArray->GetDepth(i),
                       0);
         }
         else {
            DArray->SetPos(j++,SArray->GetPos(i));
         }
      }
      matches=0;
      j--;
      for (i=0; i<=j; i++)
         if (ComparePos(DArray->GetPos(i),pattern)==0)
            break;
      for (; j>i; j--)
         if (ComparePos(DArray->GetPos(j),pattern)==0)
            break;
      for (i; i<=j; i++) {
         if (print && report)
            printf("Match (%s) at %d\n",pattern,DArray->GetPos(i));
         else if (!report) {
            matches = 1;
            if (print)
               printf("Pattern (%s) found\n",pattern);
            break;          
	 }
         matches++;
      }
      DArray->DeletePosArray();
      delete DArray;
      if (matches == 0 && !report && print)
         printf("Pattern (%s) not found\n",pattern);
      return matches;
   }
   else if (Lraja==Rraja) {
      // an occurrence found
      if (ComparePos(SArray->GetPos(Lraja),pattern)==0) {
         if (print)
            printf("Match (%s) at %d\n",pattern,SArray->GetPos(Lraja));
         return 1;
      }
   }
   if (!report && print)
         printf("Pattern (%s) not found\n",pattern);
   return 0;
}

unsigned SearchFromCSACount(char *pattern, TSA *SArray, unsigned length)
{
   unsigned k;
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   unsigned limit;
   if (SArray->GetB(i))
      i--;
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (L<R-1 && i!=L) {
         if (ComparePos(SArray->GetPos(i),pattern) <= 0)
            R = i;
         else
            L = i;
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Lraja=L+1;
         if (L==R-3 && ComparePos(SArray->GetPos(R-1),pattern) > 0)
            Lraja = R;
      }
      else
         Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   if (SArray->GetB(i))
      i--;
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1 && i!=L) {
         if (ComparePos(SArray->GetPos(i),pattern) >= 0)
            L = i;
         else
            R = i;
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Rraja=L+1;
         if (L== R-3 && ComparePos(SArray->GetPos(R-1),pattern) >= 0)
            Rraja = R-1;
      }
      else
         Rraja = L;
   }
   if ((Lraja==Rraja && SArray->GetB(Lraja)) || Lraja < Rraja) {
      TSA *DArray = new TSA();
      DArray->CreatePosArray(SArray->posbits,C);
      unsigned j=0;
      if (SArray->GetB(Lraja)) {
         UnCompactLink(SArray,
                       Lraja,
                       0,
                       DArray,
                       j,
                       SArray->GetDepth(Lraja),
                       0);
	 // search for the start of the occurrences	      
         j--;
         for (i=0; i<=j; i++)
            if (ComparePos(DArray->GetPos(i),pattern)==0)
               break;
	 matches += SArray->GetDepth(Lraja)-i;	 
      } else matches++;	 
      for (i=Lraja+1;i<Rraja;i++)
         if (SArray->GetB(i))
	    matches += SArray->GetDepth(i);
	 else 
	    matches++;
      j = 0;
      if (SArray->GetB(Rraja)) {
         UnCompactLink(SArray,
                       Rraja,
                       0,
                       DArray,
                       j,
                       SArray->GetDepth(Rraja),
                       0);
	 // search for the end of the occurrences	      
         j--;
         for (i=j; i!=0; i--)
            if (ComparePos(DArray->GetPos(i),pattern)==0)
               break;
	 matches += i+1;	 
      } else matches++;	 
      DArray->DeletePosArray();
      delete DArray;
      return matches;
   }
   else if (Lraja==Rraja) {
      // an occurrence found
      if (ComparePos(SArray->GetPos(Lraja),pattern)==0) {
         matches = 1;
      }      
   }
   return 0;
}

unsigned SearchFromSA(char *pattern, TSA *SArray, unsigned length)
{
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (L<R-1) {
         if (ComparePos(SArray->GetPos(i),pattern) <= 0)
            R = i;
         else
            L = i;
         i = (R+L)/2;
      }
      Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1) {
         if (ComparePos(SArray->GetPos(i),pattern) >= 0)
            L = i;
         else
            R = i;
         i = (R+L)/2;
      }
      Rraja=L;
   }
   matches=0;
   if (!report) {
      if (Lraja <= Rraja) {
         if (print) printf("Pattern (%s) found\n",pattern);
         return 1;
      }
      else {
         if (print) printf("Pattern (%s) not found\n",pattern);
         return 0;
      }
   }
   for (i=Lraja; i<= Rraja; i++) {
      matches++;
      if (print)
         printf("Match (%s) at %d\n",pattern,SArray->GetPos(i));
   }
   return matches;
}

unsigned SearchFromSACount(char *pattern, TSA *SArray, unsigned length)
{
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (L<R-1) {
         if (ComparePos(SArray->GetPos(i),pattern) <= 0)
            R = i;
         else
            L = i;
         i = (R+L)/2;
      }
      Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1) {
         if (ComparePos(SArray->GetPos(i),pattern) >= 0)
            L = i;
         else
            R = i;
         i = (R+L)/2;
      }
      Rraja=L;
   }
   return maxx(0,(int)(Rraja-Lraja+1));
}


unsigned SearchFromCSAWithLCP(char *pattern, TSA *SArray, unsigned length)
{
   unsigned k;
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   unsigned limit;
   unsigned lLCP, rLCP;
   unsigned mm=0;
   unsigned LCPparent,lLCPthis,rLCPthis;
   lLCP = LCP(text,SArray->GetPos(L),pattern,0);
   rLCP = LCP(text,SArray->GetPos(R),pattern,0);
   if (SArray->GetLCP(L) > SArray->GetLCP(R))
      LCPparent = SArray->GetLCP(L);
   else
      LCPparent = SArray->GetLCP(R);
   if (SArray->GetB(i))
      i--;
   assert(!SArray->GetB(i));
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (R>L+1 && i!=L) {
         if (SArray->GetLCPl(i)) {
            rLCPthis = LCPparent;
            lLCPthis = SArray->GetLCP(i);
         }
         else {
            lLCPthis = LCPparent;
            rLCPthis = SArray->GetLCP(i);
         }
         if (lLCP >= rLCP)
            if (lLCPthis >= lLCP)
               mm = lLCP+LCP(text,SArray->GetPos(i)+lLCP,pattern,lLCP);
            else
               mm = lLCPthis;
         else
            if (rLCPthis >= rLCP)
               mm = rLCP+LCP(text,SArray->GetPos(i)+rLCP,pattern,rLCP);
            else
               mm = rLCPthis;
         if (mm==m || pattern[mm] <= text[SArray->GetPos(i)+mm]) {
            R = i;
            rLCP = mm;
            LCPparent = lLCPthis;
         }
         else {
            L = i;
            lLCP = mm;
            LCPparent = rLCPthis;
         }
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Lraja=L+1;
         if (L==R-3 && ComparePos(SArray->GetPos(R-1),pattern) > 0)
            Lraja = R;
      }
      else
         Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   lLCP = LCP(text,SArray->GetPos(L),pattern,0);
   rLCP = LCP(text,SArray->GetPos(R),pattern,0);
   if (SArray->GetLCP(L) > SArray->GetLCP(R))
      LCPparent = SArray->GetLCP(L);
   else
      LCPparent = SArray->GetLCP(R);
   if (SArray->GetB(i))
      i--;
   assert(!SArray->GetB(i));
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1 && i!=L) {
         if (SArray->GetLCPl(i)) {
            rLCPthis = LCPparent;
            lLCPthis = SArray->GetLCP(i);
         }
         else {
            lLCPthis = LCPparent;
            rLCPthis = SArray->GetLCP(i);
         }
         if (lLCP >= rLCP)
            if (lLCPthis >= lLCP)
               mm = lLCP+LCP(text,SArray->GetPos(i)+lLCP,pattern,lLCP);
            else
               mm = lLCPthis;
         else
            if (rLCPthis >= rLCP)
               mm = rLCP+LCP(text,SArray->GetPos(i)+rLCP,pattern,rLCP);
            else
               mm = rLCPthis;
         if (mm==m || pattern[mm] >= text[SArray->GetPos(i)+mm]) {
            L = i;
            lLCP = mm;
            LCPparent = rLCPthis;
         }
         else {
            R = i;
            rLCP = mm;
            LCPparent = lLCPthis;
         }
         i = (R+L)/2;
         if (SArray->GetB(i))
            i--;
      }
      if (i==L) {
         Rraja=L+1;
         if (L== R-3 && ComparePos(SArray->GetPos(R-1),pattern) >= 0)
            Rraja = R-1;
      }
      else
         Rraja = L;
   }
   if (Lraja < Rraja && !report) {
      if (print) printf("Pattern (%s) found\n",pattern);
      return 1;
   }
   else if ((Lraja==Rraja && SArray->GetB(Lraja)) || Lraja < Rraja) {
      TSA *DArray = new TSA();
      DArray->CreatePosArray(SArray->posbits,(Rraja-Lraja+1)*C);
      unsigned linkstart=0;
      unsigned j=0;
      for (i=Lraja; i<=Rraja; i++) {
         if (SArray->GetB(i)) {
            linkstart=0;
            UnCompactLink(SArray,
                       i,
                       0,
                       DArray,
                       j,
                       SArray->GetDepth(i),
                       0);
         }
         else {
            DArray->SetPos(j++,SArray->GetPos(i));
         }
      }
      matches=0;
      j--;
      for (i=0; i<=j; i++)
         if (ComparePos(DArray->GetPos(i),pattern)==0)
            break;
      for (; j>i; j--)
         if (ComparePos(DArray->GetPos(j),pattern)==0)
            break;
      for (i; i<=j; i++) {
         if (print && report)
            printf("Match (%s) at %d\n",pattern,DArray->GetPos(i));
         else if (!report) {
            matches = 1;
            if (print)
               printf("Pattern (%s) found\n",pattern);
            break;          }
         matches++;
      }
      delete DArray;
      if (matches == 0 && !report && print)
         printf("Pattern (%s) not found\n",pattern);
      return matches;
   }
   else if (Lraja==Rraja) {
      // an occurrence found
      if (ComparePos(SArray->GetPos(Lraja),pattern)==0) {
         if (print)
            printf("Match (%s) at %d\n",pattern,SArray->GetPos(Lraja));
         return 1;
      }
   }
   if (!report && print)
         printf("Pattern (%s) not found\n",pattern);
   return 0;
}


unsigned SearchFromSAWithLCP(char *pattern, TSA *SArray,unsigned length)
{
   unsigned L=0;
   unsigned R=length-1;
   unsigned i=(L+R)/2;
   unsigned Lraja=0;
   unsigned Rraja=0;
   unsigned matches=0;
   unsigned lLCP, rLCP;
   unsigned mm=0;
   unsigned LCPparent;
   unsigned lLCPthis;
   unsigned rLCPthis;
   lLCP = LCP(text,SArray->GetPos(L),pattern,0);
   rLCP = LCP(text,SArray->GetPos(R),pattern,0);
   if (SArray->GetLCP(L) > SArray->GetLCP(R))
      LCPparent = SArray->GetLCP(L);
   else
      LCPparent = SArray->GetLCP(R);
   // Search for the left boundary
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Lraja=0;
   else {
      while (R>L+1 && i!=L) {
         if (SArray->GetLCPl(i)) {
            rLCPthis = LCPparent;
            lLCPthis = SArray->GetLCP(i);
         }
         else {
            lLCPthis = LCPparent;
            rLCPthis = SArray->GetLCP(i);
         }
         if (lLCP >= rLCP)
            if (lLCPthis >= lLCP)
               mm = lLCP+LCP(text,SArray->GetPos(i)+lLCP,pattern,lLCP);
            else
               mm = lLCPthis;
         else
            if (rLCPthis >= rLCP)
               mm = rLCP+LCP(text,SArray->GetPos(i)+rLCP,pattern,rLCP);
            else
               mm = rLCPthis;
         if (mm==m || pattern[mm] <= text[SArray->GetPos(i)+mm]) {
            R = i;
            rLCP = mm;
            LCPparent = lLCPthis;
         }
         else {
            L = i;
            lLCP = mm;
            LCPparent = rLCPthis;
         }
         i = (R+L)/2;
      }
      Lraja=R;
   }
   // Search for the right boundary
   L=0;
   R=length-1;
   i=(L+R)/2;
   lLCP = LCP(text,SArray->GetPos(L),pattern,0);
   rLCP = LCP(text,SArray->GetPos(R),pattern,0);
   if (SArray->GetLCP(L) > SArray->GetLCP(R))
      LCPparent = SArray->GetLCP(L);
   else
      LCPparent = SArray->GetLCP(R);
   if (ComparePos(SArray->GetPos(0),pattern) <= 0)
      Rraja=0;
   else {
      while (L < R-1 && i!=L) {
         if (SArray->GetLCPl(i)) {
            rLCPthis = LCPparent;
            lLCPthis = SArray->GetLCP(i);
         }
         else {
            lLCPthis = LCPparent;
            rLCPthis = SArray->GetLCP(i);
         }
         if (lLCP >= rLCP)
            if (lLCPthis >= lLCP)
               mm = lLCP+LCP(text,SArray->GetPos(i)+lLCP,pattern,lLCP);
            else
               mm = lLCPthis;
         else
            if (rLCPthis >= rLCP)
               mm = rLCP+LCP(text,SArray->GetPos(i)+rLCP,pattern,rLCP);
            else
               mm = rLCPthis;
         if (mm==m || pattern[mm] >= text[SArray->GetPos(i)+mm]) {
            L = i;
            lLCP = mm;
            LCPparent = rLCPthis;
         }
         else {
            R = i;
            rLCP = mm;
            LCPparent = lLCPthis;
         }
         i = (R+L)/2;
      }
         Rraja = L;
   }
   if (!report) {
      if (Lraja <= Rraja) {
         if (print) printf("Pattern (%s) found\n",pattern);
         return 1;
      }
      else {
         if (print) printf("Pattern (%s) not found\n",pattern);
         return 0;
      }
   }
   for (i=Lraja; i<= Rraja; i++) {
      matches++;
      if (print)
         printf("Match (%s) at %d\n",pattern,SArray->GetPos(i));
   }
   return matches;
}

unsigned MarkLinksToSuffixArray(TSA *SArray, TSA *DArray, unsigned D)
{
   unsigned newsize=0;
   unsigned i=0;
   unsigned j;
   unsigned x;
   unsigned char *pathdepth=(unsigned char *)malloc(n*sizeof(unsigned char));
   unsigned index;
   unsigned minLCP;
   newsize=n;
   for (i=0; i<n; i++) {
      DArray->SetPos(SArray->GetPos(i),i);
      DArray->SetB(i,0);
      SArray->SetB(i,0);
      pathdepth[i]=0;
   }
   for (i=0; i<n; i++) {
      x = DArray->GetPos(i);
      DArray->SetPos(i,SArray->GetPos(i));
      SArray->SetPos(i,x);
   }
   i = 0;
   for (i=1; i<n-2; i++) {
      if (DArray->GetPos(i) == 0)
         x = n+1;
      else
         x = SArray->GetPos(DArray->GetPos(i)-1);
      if (x < n-1 &&
          DArray->GetPos(i+1) > 0 &&
          SArray->GetPos(DArray->GetPos(i+1)-1) == x+1 &&
          (D > 255 || (pathdepth[i] + pathdepth[x] < D &&
          pathdepth[i+1] + pathdepth[x+1] < D))) {
         if (D < 256 && pathdepth[x] == 0) {
            pathdepth[x] = pathdepth[i]+1;
            pathdepth[SArray->GetPos(DArray->GetPos(i)-1+pathdepth[x])] = pathdepth[x];
            if (pathdepth[x+1]+pathdepth[i+1]+1>D) {
	       pathdepth[x]=0;
               pathdepth[SArray->GetPos(DArray->GetPos(i)-1+pathdepth[x])] =
	       pathdepth[i];
	       i++;
	       continue;
	    }
         }
         else if (D < 256) {
            pathdepth[SArray->GetPos(DArray->GetPos(i)-1-pathdepth[x])] =
              pathdepth[x]+ pathdepth[i] + 1;
            pathdepth[SArray->GetPos(DArray->GetPos(i)+pathdepth[i])] =
              pathdepth[x]+ pathdepth[i] + 1;
            if (pathdepth[x+1]+pathdepth[i+1]+1>D) {
	       pathdepth[SArray->GetPos(DArray->GetPos(i)-1-pathdepth[x])]=pathdepth[x];
               pathdepth[SArray->GetPos(DArray->GetPos(i)+pathdepth[i])] = pathdepth[i];
	       i++;
	       continue;
	    }
         }
         DArray->SetB(i,1);
         DArray->SetPos(i,x);
         if (D < 256 && pathdepth[x+1] == 0) {
            pathdepth[x+1] = pathdepth[i+1]+1;
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)-1+pathdepth[x+1])] =
            pathdepth[x+1];
         }
         else if (D<256) {
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)-1-pathdepth[x+1])] =
              pathdepth[x+1]+ pathdepth[i+1] + 1;
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)+pathdepth[i+1])] =
              pathdepth[x+1]+ pathdepth[i+1] + 1;
         }
         DArray->SetB(i+1,1);
         DArray->SetPos(i+1,x+1);
         newsize--;
         j = 1;
         while (j+1 < C && x+j+1 <n && i+j+1 <n-1 &&
                (D>255 || (pathdepth[i+j+1] + pathdepth[x+j+1] < D)) &&
                DArray->GetPos(i+j+1) > 0 &&
                SArray->GetPos(DArray->GetPos(i+j+1)-1) == x+j+1) {
            if (D <256 && pathdepth[x+j+1] == 0) {
               pathdepth[x+j+1] = pathdepth[i+j+1]+1;
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)-1+pathdepth[x+j+1])] =
               pathdepth[x+j+1];
            }
            else if (D<256) {
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)-1-pathdepth[x+j+1])] =
                 pathdepth[x+j+1]+ pathdepth[i+j+1] + 1;
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)+pathdepth[i+j+1])] =
                 pathdepth[x+j+1]+ pathdepth[i+j+1] + 1;
            }
            DArray->SetB(i+j+1,1);
            DArray->SetPos(i+j+1,x+j+1);
            newsize--;
            j++;
         }
         i = i+j+1;
      }
   }
   for (i=0; i<n; i++) {
      j = SArray->GetPos(i);
      x = i;
      while (SArray->GetB(j)==0) {
         index = j;
         j = SArray->GetPos(index);
         SArray->SetPos(index,x);
         SArray->SetB(index,1);
         x = index;
      }
   }
   for (i=0; i<n; i++)
     assert(SArray->GetPos(i)<n);
   return newsize;
}

unsigned CalcNewSize(TSA *SArray, unsigned D)
{
   unsigned newsize=0;
   unsigned i=0;
   unsigned j;
   unsigned x;
   unsigned char *pathdepth=(unsigned char *)malloc(n*sizeof(unsigned char));
   TSA *DArray= new TSA();
   DArray->CreatePosArray(bitsn,n+1);
   unsigned index;
   newsize=n;
   for (i=0; i<n; i++) {
      DArray->SetPos(SArray->GetPos(i),i);
      DArray->SetB(i,0);
      SArray->SetB(i,0);
      pathdepth[i]=0;
   }
   for (i=0; i<n; i++) {
      x = DArray->GetPos(i);
      DArray->SetPos(i,SArray->GetPos(i));
      SArray->SetPos(i,x);
   }
   i = 0;
   for (i=1; i<n-1; i++) {
      if (DArray->GetPos(i) == 0)
         x = n+1;
      else
         x = SArray->GetPos(DArray->GetPos(i)-1);
      if (x < n-1 && pathdepth[i] + pathdepth[x] < D &&
          pathdepth[i+1] + pathdepth[x+1] < D && DArray->GetPos(i+1) > 0 &&
          SArray->GetPos(DArray->GetPos(i+1)-1) == x+1) {
         if (pathdepth[x] == 0) {
            pathdepth[x] = pathdepth[i]+1;
            pathdepth[SArray->GetPos(DArray->GetPos(i)-1+pathdepth[x])] = pathdepth[x];
         }
         else {
            pathdepth[SArray->GetPos(DArray->GetPos(i)-1-pathdepth[x])] =
              pathdepth[x]+ pathdepth[i] + 1;
            pathdepth[SArray->GetPos(DArray->GetPos(i)+pathdepth[i])] =
              pathdepth[x]+ pathdepth[i] + 1;
         }
         if (pathdepth[x+1] == 0) {
            pathdepth[x+1] = pathdepth[i+1]+1;
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)-1+pathdepth[x+1])] =
            pathdepth[x+1];
         }
         else {
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)-1-pathdepth[x+1])] =
              pathdepth[x+1]+ pathdepth[i+1] + 1;
            pathdepth[SArray->GetPos(DArray->GetPos(i+1)+pathdepth[i+1])] =
              pathdepth[x+1]+ pathdepth[i+1] + 1;
         }
         newsize--;
         j = 1;
         while (j+1 < C && x+j+1 <n && i+j+1 <n-1 &&
                pathdepth[i+j+1] + pathdepth[x+j+1] < D &&
                DArray->GetPos(i+j+1) > 0 &&
                SArray->GetPos(DArray->GetPos(i+j+1)-1) == x+j+1) {
            if (pathdepth[x+j+1] == 0) {
               pathdepth[x+j+1] = pathdepth[i+j+1]+1;
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)-1+pathdepth[x+j+1])] =
               pathdepth[x+j+1];
            }
            else {
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)-1-pathdepth[x+j+1])] =
                 pathdepth[x+j+1]+ pathdepth[i+j+1] + 1;
               pathdepth[SArray->GetPos(DArray->GetPos(i+j+1)+pathdepth[i+j+1])] =
                 pathdepth[x+j+1]+ pathdepth[i+j+1] + 1;
            }
            newsize--;
            j++;
         }
         i = i+j+1;
      }
   }
   for (i=0; i<n; i++) {
      j = SArray->GetPos(i);
      x = i;
      while (SArray->GetB(j)==0) {
         index = j;
         j = SArray->GetPos(index);
         SArray->SetPos(index,x);
         SArray->SetB(index,1);
         x = index;
      }
   }
   return newsize;
}

void CreateCompactedSuffixArray(TSA *SArray, TSA *DArray,int newsize)
{
   unsigned i;
   unsigned j=0;
   unsigned x=0;
   SArray->SetB(0,0);
   for (i=0;i<n;i++)
      if (SArray->GetB(i)==0) {
         DArray->SetPos(j,SArray->GetPos(i));
         DArray->SetB(j,0);
         SArray->SetPos(i,j++);
      }
      else {
         DArray->SetPos(j,SArray->GetPos(i));
         DArray->SetB(j,1);
         DArray->SetDepth(j,1);
         SArray->SetPos(i,j++);
         SArray->SetB(i++,0);
         while (SArray->GetB(i)) {
            x = DArray->GetDepth(j-1);
            SArray->SetPos(i++,x);
            DArray->SetDepth(j-1,x+1);
         }
         DArray->SetPos(j,SArray->GetPos(i));
         DArray->SetB(j,0);
         SArray->SetPos(i,j++);
         SArray->SetB(i,0);
      }
   for (i=0; i<newsize; i++)
         if (DArray->GetB(i) && SArray->GetB(DArray->GetPos(i)) == 0) {
            DArray->SetPos(i, SArray->GetPos(DArray->GetPos(i)));
            DArray->SetJump(i,0);
         }
         else if (DArray->GetB(i)) {
            DArray->SetJump(i,SArray->GetPos(DArray->GetPos(i)));
            DArray->SetPos(i,SArray->GetPos(DArray->GetPos(i)-DArray->GetJump(i)));
         }
}


void CopyAdjacentLCPsToCSA(TSA *SArray,TSA *CArray)
{
   unsigned i,j=0, minLCP;
   for (i=0; i<n; i++) {
      if (SArray->GetB(i) == 0)
         CArray->SetLCP(j++,minn(maxLCP,SArray->GetLCP(i)));
      else {
         j++;
         minLCP = SArray->GetLCP(i++);
         while (SArray->GetB(i)) {
            if (SArray->GetLCP(i) < minLCP)
               minLCP = SArray->GetLCP(i);
            i++;
         }
         if (SArray->GetLCP(i) < minLCP)
            CArray->SetLCP(j++,minn(maxLCP,SArray->GetLCP(i)));
         else
            CArray->SetLCP(j++,minn(maxLCP,minLCP));
      }
   }
}


double CompactSuffixArray(int alg, unsigned D)
{
   unsigned newsize=n;
   if (C==0) {
      unsigned i, minC=0;
      float minsize=n*(ceil(log(1.0*n)/log(2.0))+1)/8;
      for (i=1; i<16; i++) {
         C = (1 << i)-1;
         newsize = CalcNewSize(SA, D);
         printf("C=%d, newsize=%d\n", C, newsize);
         if (newsize*((ceil(log(1.0*n)/log(2.0))+1)/8+1.0*i/8) < minsize) {
            minsize = newsize*((ceil(log(1.0*n)/log(2.0))+1)/8+1.0*i/8);
            minC = C;
         }
      }
      C = minC;
      maxLCP = (C+1)*(C+1)/2-1;
   }
   TSA *SATemp = new TSA();
   SATemp->CreatePosArray(bitsn,n+1);
   newsize = MarkLinksToSuffixArray(SA, SATemp, D);
   if (!savemem && (alg == 2 || alg == 5)) {
      SA->CreateLCPArray(bitsn+1,n+1,false);
      CalcAdjacentLCPsSA(SA);
   }
   SA->DeletePosArray();
   CSA = new TSA();
   unsigned bitsd;
   if (alg == 5) {
      maxLCP = (1u << ((unsigned)ceil(log(1.0*C)/log(2.0))*2-1))-1;
      bitsd = (unsigned)ceil(log(1.0*maxx((C+1)*(C+1),(maxLCP+1)*2))/log(2.0));
   }
   else {
      bitsd = (unsigned)ceil(log(1.0*maxx((C+1),(maxLCP+1)*2))/log(2.0));
   }
   bool shortarray=true;
   if (alg==2 || alg==5)
      shortarray=false;
   CSA->CreateLCPArray(bitsd,newsize+1,shortarray);
   if (!savemem && (alg == 2 || alg == 5)) {
      SATemp->lcp = SA->lcp;
      SATemp->lcpbits = SA->lcpbits;
      CopyAdjacentLCPsToCSA(SATemp,CSA);
      SA->DeleteLCPArray();
      delete SA;
   }
   CSA->CreatePosArray(bitsn,newsize+1);
   SuffixArrayCompactSize = newsize;
   CreateCompactedSuffixArray(SATemp,CSA,newsize);
   SATemp->DeletePosArray();
   delete SATemp;
   if (!savemem && (alg == 2 || alg == 5))
      CalcLCPsFromAdjacentLCPs(CSA,0,newsize-1);
   else if (alg == 2 || alg == 5)
      CalcLCPs(CSA,0,newsize-1);

	double cantBytesCSA = newsize*((1.0*CSA->posbits+1)/8+1.0*CSA->lcpbits/8);
	double porcent = (double)cantBytesCSA / (double)n;

	printf("Compacted array without BitSequebceRRR Lmap takes %d*%f B = %f B space = %f|Text| \n",
          newsize,(1.0*CSA->posbits+1)/8+
          1.0*CSA->lcpbits/8,
          cantBytesCSA, porcent);
	fflush(stdout);

	return cantBytesCSA;
}


void UnCompactLink(TSA *CArray,unsigned Cindex, unsigned depth, TSA *Area, unsigned &Areasize, unsigned remaining, unsigned startparent) {
   unsigned i=0;
   unsigned j=CArray->GetPos(Cindex);
   unsigned startthis = CArray->GetJump(Cindex)+startparent;
   unsigned thisdepth = minn(CArray->GetDepth(Cindex)-startparent,remaining);
   unsigned Areastart = Areasize;

   while (startthis > 0){
   		if (!CArray->GetB(j+i)) {
   			i++;
   			startthis--;
   			operations++;
   		}
   		else{
   			if (CArray->GetDepth(j+i) <= startthis)  {
   				startthis -= CArray->GetDepth(j+i++);
   				operations++;
   			}
   			else break;
   		}
   }

   while (1) {
      if (CArray->GetB(j+i)) {
         UnCompactLink(CArray, j+i, depth+1, Area, Areasize, thisdepth-(Areasize-Areastart), startthis);
         operations++;
         if (Areasize-Areastart>= thisdepth)
            break;
         startthis=0;
      }
      else {
         Area->SetPos(Areasize++,CArray->GetPos(j+i)+depth+1);
         sumdepth += depth+1;
         operations++;
         if (Areasize-Areastart >= thisdepth)
            break;
      }
      i++;
   }

   if (depth+1 > maxdepth)
      maxdepth = depth+1;
}


// Functions for experiments...
void UnCompactOffsetLink(TSA *CArray, unsigned Cindex, unsigned depth, unsigned &value, unsigned &Areasize, unsigned remaining, unsigned startparent, unsigned offset){
	unsigned i=0;
	unsigned j=CArray->GetPos(Cindex);
	unsigned startthis = CArray->GetJump(Cindex)+startparent;
	unsigned thisdepth = minn(CArray->GetDepth(Cindex)-startparent,remaining);
	unsigned Areastart = Areasize;

	if (Areasize > offset) return;

	while (startthis > 0){
		if (!CArray->GetB(j+i)) {
			i++;
			startthis--;
			operations++;
		}
		else{
			if (CArray->GetDepth(j+i) <= startthis)  {
				startthis -= CArray->GetDepth(j+i++);
				operations++;
			}
			else break;
		}
	}

	while (1) {
	  if (CArray->GetB(j+i)) {
		  UnCompactOffsetLink(CArray, j+i, depth+1, value, Areasize, thisdepth-(Areasize-Areastart), startthis, offset);
		  operations++;
		  if (Areasize-Areastart>= thisdepth)
			  break;
		  startthis=0;
	  }
	  else {
		  if(Areasize == offset){
			  value = CArray->GetPos(j+i)+depth+1;
			  Areasize++;
			  return;
		  }
		  else
			  Areasize++;
		  sumdepth += depth+1;
		  operations++;
		  if (Areasize-Areastart >= thisdepth)
			break;
	  }
	  i++;
	}

	if (depth+1 > maxdepth)
	  maxdepth = depth+1;
}

void UnCompactLinkExp(TSA *CArray,unsigned Cindex, unsigned depth, unsigned *items, unsigned &Areasize, unsigned remaining, unsigned startparent) {
   unsigned i=0;
   unsigned j=CArray->GetPos(Cindex);
   unsigned startthis = CArray->GetJump(Cindex)+startparent;
   unsigned thisdepth = minn(CArray->GetDepth(Cindex)-startparent,remaining);
   unsigned Areastart = Areasize;

   while (startthis > 0){
   		if (!CArray->GetB(j+i)) {
   			i++;
   			startthis--;
   			operations++;
   		}
   		else{
   			if (CArray->GetDepth(j+i) <= startthis)  {
   				startthis -= CArray->GetDepth(j+i++);
   				operations++;
   			}
   			else break;
   		}
   }

   while (1) {
      if (CArray->GetB(j+i)) {
         UnCompactLinkExp(CArray, j+i, depth+1, items, Areasize, thisdepth-(Areasize-Areastart), startthis);
         operations++;
         if (Areasize-Areastart>= thisdepth)
            break;
         startthis=0;
      }
      else {
    	  items[Areasize++] = CArray->GetPos(j+i)+depth+1;
    	  sumdepth += depth+1;
    	  operations++;
    	  if (Areasize-Areastart >= thisdepth)
    		  break;
      }
      i++;
   }

   if (depth+1 > maxdepth)
      maxdepth = depth+1;
}

// extract the firsts offset cells
void UnCompactDeltaLink(TSA *CArray, unsigned Cindex, unsigned depth, unsigned *items, unsigned &Areasize, unsigned remaining, unsigned startparent, unsigned &offset){
	unsigned i=0;
	unsigned j=CArray->GetPos(Cindex);
	unsigned startthis = CArray->GetJump(Cindex)+startparent;
	unsigned thisdepth = minn(CArray->GetDepth(Cindex)-startparent,remaining);
	unsigned Areastart = Areasize;

	while (startthis > 0){
   		if (!CArray->GetB(j+i)) {
   			i++;
   			startthis--;
   			operations++;
   		}
   		else{
   			if (CArray->GetDepth(j+i) <= startthis)  {
   				startthis -= CArray->GetDepth(j+i++);
   				operations++;
   			}
   			else break;
   		}
   	}

   	while (1) {
	   if (CArray->GetB(j+i)) {
		   UnCompactDeltaLink(CArray, j+i, depth+1, items, Areasize, thisdepth-(Areasize-Areastart), startthis, offset);
		   if(offset == 0) return;
		   operations++;
		   if (Areasize-Areastart>= thisdepth)
			   break;
		   startthis=0;
	   }
	   else {
		   items[Areasize++] = CArray->GetPos(j+i)+depth+1;
		   offset--;
		   if(offset == 0) return;
		   else{
			   sumdepth += depth+1;
			   operations++;
			   if (Areasize-Areastart >= thisdepth)
				   break;
		   }
	   }
	   i++;
   	}
   	if (depth+1 > maxdepth)
   		maxdepth = depth+1;
}
/*void UnCompactDeltaLink(TSA *CArray, unsigned Cindex, unsigned depth, TSA *Area, unsigned &Areasize, unsigned remaining, unsigned startparent, unsigned &offset){
	unsigned i=0;
	unsigned j=CArray->GetPos(Cindex);
	unsigned startthis = CArray->GetJump(Cindex)+startparent;
	unsigned thisdepth = minn(CArray->GetDepth(Cindex)-startparent,remaining);
	unsigned Areastart = Areasize;

	while (startthis > 0){
   		if (!CArray->GetB(j+i)) {
   			i++;
   			startthis--;
   			operations++;
   		}
   		else{
   			if (CArray->GetDepth(j+i) <= startthis)  {
   				startthis -= CArray->GetDepth(j+i++);
   				operations++;
   			}
   			else break;
   		}
   	}

   	while (1) {
	   if (CArray->GetB(j+i)) {
		   UnCompactDeltaLink(CArray,j+i,depth+1,Area,Areasize,thisdepth-(Areasize-Areastart),startthis, offset);
		   if(offset == 0) return;
		   operations++;
		   if (Areasize-Areastart>= thisdepth)
			   break;
		   startthis=0;
	   }
	   else {
		   Area->SetPos(Areasize++,CArray->GetPos(j+i)+depth+1);
		   offset--;
		   if(offset == 0) return;
		   else{
			   sumdepth += depth+1;
			   operations++;
			   if (Areasize-Areastart >= thisdepth)
				   break;
		   }
	   }
	   i++;
   	}
   	if (depth+1 > maxdepth)
   		maxdepth = depth+1;
}*/

void UnCompactSuffixArray(unsigned D)
{
   TSA *DArray = new TSA();
   DArray->CreatePosArray(bitsn,n+1);
   unsigned linkstart=0;
   unsigned newsize;
   unsigned i, j, dd, oldj;
   dd=1;
   j=0;
   maxdepth = 0;
   unsigned links=0;
   sumdepth = 0.0;
   unsigned suffixes=0;
   operations=0;
   for (i=0; i<10000; i++)
      depthcount[i]=0;

   for (i=0; i<SuffixArrayCompactSize;i++) {
      if (CSA->GetB(i)) {
         linkstart=0;
         oldj = j;
         links++;
         //startownclock();
         operations++;
         UnCompactLink(CSA,
                       i,
                       0,
                       DArray,
                       j,
                       CSA->GetDepth(i),
                       0);
         //tuc = stopownclock();
         tuc = 0;
         if (tuc > maxtuc) {
            maxtuc = tuc;
            maxtucindex = i;
            maxtucsize = j-oldj;
         }
         if (maxdepth > D) { // !ERROR!
        	 int t;
			printf("Error. Link:\n");
			for (t=oldj;t<j;t++)
			   printf("%d ",DArray->GetPos(t));
			printf("\n");
			maxdepth = D;
         }
      }
      else {
         DArray->SetPos(j++,CSA->GetPos(i));
         suffixes++;
      }

      if (print)
         if (j*10.0/dd > n) {
            printf("%d ",10*dd++);
            fflush(stdout);
         }
   }

   double sumdepthv=0.0;
   for (i=0; i<10000; i++)
      sumdepthv +=depthcount[i]*(i-sumdepth/(n-suffixes))*(i-sumdepth/(n-suffixes));
   if (print)
      printf("\n");
   /*if (print)
      printf("Max uncompactime %4fs at index %d (jump=%d, length=%d),
      maxdepth=%d, avgdepth=%f, sqrt(variance)=%f, operations=%d, links=%d\n",
              maxtuc*1.0/hzz,maxtucindex,(int)(CSA->GetJump(maxtucindex)),maxtucsize,maxdepth,
              sumdepth/(n-suffixes),sqrt(sumdepthv/(n-suffixes)),operations,links);*/
   newsize = j;
   j=n+1;
   printf("UnCompacted array takes %d*%f B = %f B space \n",
          newsize,1.0*DArray->posbits/8,
          newsize*1.0*DArray->posbits/8);
   DArray->DeletePosArray();
   delete DArray;
}


void CreateSuffixArray(int *p, bool listTraza, bool runTest)
{
	int *x;
	unsigned i;

	x = (int*)malloc((n+1)*sizeof *x);
	for (i=0; i<n; i++)
	  x[i]=(int)text[i]+128;
	x[n]=0;

	printf("CAll to suffixsort...\n");
	fflush(stdout);
	suffixsort(x,p,n,256,0);
	printf("Return from suffixsort...\n");

	free(x);
	unsigned k=0;
	while (((n+1)<<k)>>31 == 0) k++;
	bitsn = 32-k;
	SA = new TSA();

	printf("CAll to CreatePosArray...\n");
	SA->CreatePosArray(bitsn,n+1);
	SA->lcpbits = 0;
	printf("Return from CreatePosArray...\n");

	for (i=0; i<n; i++) {
	  SA->SetPos(i,p[i+1]);
	  SA->SetB(i,0);
	}
	printf("End CreateSuffixArray!!\n");
}

double MakeSuffixArray(int alg, unsigned D, bool listTraza, bool runTest)
{
	int *p;
	double t1Construction, t2Construction, cantBytesCSA;
	unsigned i, j, depth, posL;

	printf("MakeSuffixArray Start...\n");
	fflush(stdout);
	p=(int*)malloc((n+1)*sizeof *p);
	CreateSuffixArray(p, listTraza, runTest);

	t1Construction = getTime();
	printf("CreateSuffixArray Ok...\n");
	fflush(stdout);

	if (alg == 3) {
      SA->CreateLCPArray(bitsn+1,n+1,false);
      CalcAdjacentLCPsSA(SA);
      TSA * SATemp = new TSA();
      SATemp->CreateLCPArray((unsigned)ceil(log(maxLCP+1)/log(2.0))+1,n+1,false);
      for (i=0; i<n; i++)
         SATemp->SetLCP(i,SA->GetLCP(i));
      SA->DeleteLCPArray();
      SA->lcp = SATemp->lcp;
      SA->lcpbits = SATemp->lcpbits;
      delete SATemp;
	}
	else if (savemem && (alg == 2 || alg == 5)) {
      SA->CreateLCPArray(bitsn+1,n+1,false);
      CalcAdjacentLCPsSA(SA);
      SA->DeleteLCPArray();
	} else {
      maxLCP = 0;
      SA->lcpbits = 0;
	}

	printf("Text size is %d B\n",n);
	printf("Array takes %d*%f B = %f B space = %f|Text| \n",
		  n,(1.0*SA->posbits+SA->lcpbits)/8,
		  n*((1.0*SA->posbits+SA->lcpbits)/8), (1.0*SA->posbits+SA->lcpbits)/8.0);
	fflush(stdout);

	if (alg == 0 || alg == 2 || alg == 4 || alg == 5){
		cantBytesCSA = CompactSuffixArray(alg, D);
		printf("CompactSuffixArray Ok...\n");
		fflush(stdout);
	}else if (alg == 3){
		CalcLCPsFromAdjacentLCPs(SA,0,n-1);
		printf("CalcLCPsFromAdjacentLCPs Ok...\n");
		fflush(stdout);
	}

	 // Create bitSequenceRRR
	BitString * Lstr = new BitString(n);
	for(i=0,posL=0; i<SuffixArrayCompactSize; i++){
	   if (CSA->GetB(i)) {
		   Lstr->setBit(posL,true);
		   posL++;
		   depth = CSA->GetDepth(i);
		   for(j=1; j<CSA->GetDepth(i); j++){
			   Lstr->setBit(posL,false);
			   posL++;
		   }
	   }else{
		   Lstr->setBit(posL,true);
		   posL++;
	   }
	}
	BitSequenceRRR *bsRRR = new BitSequenceRRR(*Lstr, 33);
	Lmap = saveLoad(bsRRR);
	delete (BitString *) Lstr;
	printf("BitSequenceRRR Lmap Ok...\n");
	fflush(stdout);

	t2Construction = getTime();

	printf("n = %d\n", n);
	printf("SuffixArrayCompactSize = %d\n", SuffixArrayCompactSize);
	cantBytesCSA += Lmap->getSize();
	printf("++++++++++++++++++++++++++++++++++++++++++\n");
	printf("Final CSA size = %f Bytes = %f|Text|\n", cantBytesCSA, cantBytesCSA/n);
	printf("++++++++++++++++++++++++++++++++++++++++++\n");
	fflush(stdout);



	if (listTraza){
	   PrintSuffixArray(CSA, SuffixArrayCompactSize);
	   printf("\nText = [%s]\n",text);
	   fflush(stdout);
	   printf("SA[1..%d] = \n",n);
	   for (i=1; i<=n; i++)
		   printf("%d ",p[i]);
	   fflush(stdout);

	   printf("\nBitSequenceRRR :\n");
	   for (i=0;i<n;i++)
		   printf("%d", Lmap->access(i));
	   fflush(stdout);

	   printf("\n%d", Lmap->rank1(0));
	   for (i=1;i<n;i++)
		   printf("%d", Lmap->rank1(i)-Lmap->rank1(i-1));
	   fflush(stdout);

	   printf("\nCSA->GetB[1..%d] :\n", SuffixArrayCompactSize);
	   for (i=0;i<SuffixArrayCompactSize;i++)
		   printf("%d ",CSA->GetB(i));
	   fflush(stdout);

	   printf("\nCSA->GetPos[1..%d] :\n", SuffixArrayCompactSize);
	   for (i=0;i<SuffixArrayCompactSize;i++)
		   printf("%d ",CSA->GetPos(i));
	   fflush(stdout);

	   printf("\nCSA->GetDepth[1..%d] :\n", SuffixArrayCompactSize);
	   for (i=0;i<SuffixArrayCompactSize;i++)
		   printf("%d ",CSA->GetDepth(i));
	   printf("\n");
	   fflush(stdout);
   }

   if (runTest){
	   unsigned repetitions = 10000;
	   if (repetitions > n)
		   repetitions = n;
   	   printf("_______________________________________________________________________________________________________\n");
   	   printf("Test retrieve A[1..n]...\n");
   	   for (i=0; i<repetitions; i++){
   		   if (accessOneFromSA(i) != p[i+1]){
   			   printf("****** ERROR !! ******\n   accessOneFromSA(%d) = %d != p[%d+1] = %d", i, accessOneFromSA(i), i, p[i]);
   			   exit(1);
   		   }
   	   }
   	   printf("Test retrieve A[1..n] Finished OK !!\n");

   	   printf("_______________________________________________________________________________________________________\n");
   	   printf("Test retrieve A[i..j]... \n");

   	   unsigned *items = new unsigned[1000];
   	   unsigned start, k, range = 100;
   	   if (range > n)
   		   range = 10;

   	   for (i=0; i<repetitions; i++){
   		   j = (rand() % (n-range+1));
   		   start = accessIntervalFromSA(j, j+range-1, items);
   		   for (k=0; k<range; k++){
   			   if (items[start+k] != p[j+k+1]){
   				   printf("****** ERROR in accessIntervalFromSA!! ******\n (j=%d), start=%d,  items[%d] = %d != p[%d+%d+1] = %d", j, start, start+k, items[start+k], j, k, p[j+k+1]);
   				   exit(1);
   			   }
   		   }
   	   }

   	   delete (unsigned *) items;
   	   printf("Test retrieve A[i..j] Finished OK !!\n\n");
   	   free(p);
   }

   return (t2Construction - t1Construction);
}

void PrintSuffixArray(TSA *SArray, unsigned length)
{
	printf("length=%d\n",length);
   unsigned i;
   for (i=0; i<256 && i < length; i++) {
      if (!SArray->GetB(i) && SArray->lcpbits == 0)
         printf("%d:(%d) %s \n",i,SArray->GetPos(i),SubString(text,SArray->GetPos(i),20));
      else if (!SArray->GetB(i) && SArray->GetLCPl(i))
         printf("%d:(%d) %s (l=%d)\n",i,SArray->GetPos(i),SubString(text,SArray->GetPos(i),20),
                (int)SArray->GetLCP(i));
      else if (!SArray->GetB(i) && SArray->GetLCPr(i))
         printf("%d:(%d) %s (r=%d)\n",i,SArray->GetPos(i),SubString(text,SArray->GetPos(i),20),
                (int)SArray->GetLCP(i));
      else
         printf("%d:->%d (%d) (%d) \n",i,SArray->GetPos(i),
                (int)SArray->GetDepth(i),(int)SArray->GetJump(i));
      fflush(stdout);
   }
}

// this is used for libcds for manipulate bits sequences
BitSequence * saveLoad(BitSequence * bs) {
    ofstream ofs("bitsequence.tmp");
    bs->save(ofs);
    ofs.close();
    ifstream ifs("bitsequence.tmp");
    BitSequence * ret = BitSequence::load(ifs);
    ifs.close();
    return ret;
}

double getTime (void){
	double usertime, systime;
	struct rusage usage;

	getrusage (RUSAGE_SELF, &usage);

	usertime = (double) usage.ru_utime.tv_sec +
	(double) usage.ru_utime.tv_usec / 1000000.0;

	systime = (double) usage.ru_stime.tv_sec +
	(double) usage.ru_stime.tv_usec / 1000000.0;

	return (usertime + systime);
}

// retrieve A[pos]
unsigned accessOneFromSA(unsigned pos){
	unsigned posCSA = Lmap->rank1(pos)-1;
	unsigned value=0;

	if (CSA->GetB(posCSA)){		// is it marked this position ?
		int offset = pos - Lmap->select1(posCSA+1);
		unsigned j=0;

		UnCompactOffsetLink(CSA, posCSA, 0, value, j, CSA->GetDepth(posCSA), 0, offset);
	}else
		value = CSA->GetPos(posCSA);

	return value;
}

// Experiment 1 - compute A[i]: it gives the average time of CPU
double runExperimentOne(unsigned repetitions){
	double t, avgTime = 0.0;
	unsigned i, j, value;

	printf("Start Experiment 1,  Extract A[i], for %d repetitions\n",repetitions);

	for (i=0; i<repetitions; i++){
		j = (rand() % n);
		t = getTime();
		value = accessOneFromSA(j);
		t = getTime() - t;
		avgTime += t/(double)repetitions;
		//printf("A[%d] = %d\n",j,value);
	}

	printf("Average CPU time for item : %f Milliseconds\n", avgTime * 1000.0);
	printf("Experiment 1 has finished\n");
	printf("_______________________________________________________________________________________________________\n");

	return avgTime;
}

// retrieve A[i,j]
unsigned accessIntervalFromSA(unsigned i, unsigned j, unsigned *items){
	unsigned left = Lmap->rank1(i)-1;
	unsigned currentIndexNewArea = 0;
	unsigned currentIndexCSA = left;
	unsigned start = i - Lmap->select1(left+1);
	unsigned interval = j-i+1+start;
	unsigned delta;

	// Extract CSA[left..right]=CSA[left..left+offset] into items[]
	for (currentIndexCSA = left; currentIndexNewArea < interval; currentIndexCSA++) {
		if (CSA->GetB(currentIndexCSA)) {
			if ((currentIndexNewArea + CSA->GetDepth(currentIndexCSA)) > interval){
				delta = interval - currentIndexNewArea;
				UnCompactDeltaLink(CSA, currentIndexCSA, 0, items, currentIndexNewArea, CSA->GetDepth(currentIndexCSA), 0, delta);
			}else
				UnCompactLinkExp(CSA, currentIndexCSA, 0, items, currentIndexNewArea, CSA->GetDepth(currentIndexCSA), 0);

		}else{
			items[currentIndexNewArea] = CSA->GetPos(currentIndexCSA);
			currentIndexNewArea++;
		}
	}

	return start;
}

// Experiment 2 - compute A[i..j]: it gives the average time of CPU
double runExperimentTwo(unsigned repetitions, unsigned itemsExp2, unsigned *items){
	double t, avgTime = 0.0;
	unsigned i, j;

	printf("Start Experiment 2. Extract A[i..i+%d], for %d repetitions\n", itemsExp2, repetitions);

	for (i=0; i<repetitions; i++){
		j = (rand() % (n-itemsExp2+1));
		t = getTime();
		accessIntervalFromSA(j, j+itemsExp2-1, items);
		t = getTime() - t;
		avgTime += t/(double)repetitions;
	}
	t = getTime() - t;
	printf("Average CPU time for item : %f Milliseconds\n", avgTime * 1000.0);
	printf("Experiment 2 has finished\n");
	printf("_______________________________________________________________________________________________________\n");

	return avgTime;
}

// Experiment 3 - compute A[i..i+1], A[i..i+2], ..., A[i..i+%d], it gives the average time of CPU
double runExperimentThree(unsigned repetitions, unsigned itemsExp3, unsigned *items, double * timesItems){
	double t;
	unsigned r, i, j;

	for (i=0; i<itemsExp3; i++)
		timesItems[i] = 0.0;

	printf("Start Experiment 3. Extract A[i], A[i..i+1], A[i..i+2], ..., A[i..i+%d], for %d repetitions\n", itemsExp3-1, repetitions);
	for (i=0; i<itemsExp3; i++){
		for (r=0; r<repetitions; r++){
			j = (rand() % (n-itemsExp3+1));
			t = getTime();
			accessIntervalFromSA(j, j+i, items);
			t = getTime() - t;
			timesItems[i] = timesItems[i] + t/(double)repetitions;
		}
	}

	printf("CPU time for each c = {1,2,...,%d} in Milliseconds\n", itemsExp3);
	for (i=0; i<itemsExp3; i++)
		printf("%f ", timesItems[i]*1000.0);

	printf("\nExperiment 3 has finished\n");
	printf("_______________________________________________________________________________________________________\n");

	return 0;
}


void runExperiments(unsigned itemsExp2, unsigned itemsExp3){
	unsigned *items = new unsigned[1000];
	double *timesItems = new double[itemsExp3];

	unsigned repetitions = 1000;

	runExperimentOne(repetitions);									// 1000000 repetitions
	runExperimentTwo(600000, itemsExp2, items);					//  600000 repetitions
	if (n>200)
		runExperimentTwo(200000, 100, items);					//  200000 repetitions for 100 continuous items

	runExperimentThree(repetitions, itemsExp3, items, timesItems);	// 1000000 repetitions

	delete (unsigned *) items;
	delete (double *) timesItems;
}

// to calculate stdev:
/*for(repetitions=100000; repetitions<=1000000; repetitions+=100000)
	runExperimentTwo(repetitions, 100, items);
for(repetitions=10000; repetitions<=100000; repetitions+=10000)
	runExperimentThree(repetitions, itemsExp3, items, timesItems);*/
